#!/usr/bin/env python
# coding: utf-8


# Import all libraries
import torch
from torch import nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torch.utils.data import Subset
from torchvision.transforms import ToTensor
from torchmetrics import Accuracy
from torchvision.datasets import MNIST
from torchvision import transforms
import pytorch_lightning as pl
from pytorch_lightning import LightningModule, Trainer
import numpy as np
from sklearn.model_selection import KFold
import pandas as pd
import time
import matplotlib.pyplot as plt
from pynvml import *
import cpuinfo
import os

pl.seed_everything(42, workers=True)
torch.manual_seed(42)

# Define variables that should be augmented for benchmarking purposes

# Obtain the number of available GPUs and adjust the CPU variable accordingly
# Then determine the appropriate batch size given the computational resources
number_of_gpus = torch.cuda.device_count()
print('number_of_gpus:')
print(number_of_gpus)
print('')
batch_size = 256 if number_of_gpus >= 1 else 64

# Record the GPU type (if one is present)
if number_of_gpus == 0:
    gpu_type = "None"
else:
    nvmlInit()
    gpu_type = nvmlDeviceGetName(nvmlDeviceGetHandleByIndex(0))

# Specify the number of cores to use, adjusting per GPU usage
number_of_cores = int(sys.argv[1])
number_of_cores = 1 if number_of_gpus>=1 else number_of_cores

# Input k for K-Fold CV
k = int(sys.argv[2])

# Input the number of images to keep from the full dataset
n_images = int(sys.argv[3])

# Use a string for file organization purposes
output_string = sys.argv[4]

# Record the processor platform and output all of the relevant data that was produced
cpu_type = cpuinfo.get_cpu_info().get("brand_raw")
cpu_speed = cpuinfo.get_cpu_info().get("hz_advertised_friendly")
cpu_count = number_of_cores


# Produce the folds and load the raw data
kf_outer = KFold(n_splits=k, shuffle=True, random_state=42)
kf_inner = KFold(n_splits=k-1, shuffle=True, random_state=42)
raw_data = MNIST(root=".", train=True, download=True, transform=ToTensor())
selected_data = Subset(raw_data, range(n_images))

# Make a new log directory using metadata from the CLI
file_name = output_string+"_"+str(cpu_count)+"cpus_"+str(number_of_gpus)+"gpus_"+str(k)+"k_"+str(n_images).zfill(5)+"images_"+time.strftime("%Y%m%d%H%M%S")
directory = file_name
parent_dir = "./lightning_logs"
log_path = os.path.join(parent_dir, directory) 
os.makedirs(parent_dir,exist_ok=True) 
os.makedirs(log_path,exist_ok=True) 


# Make a relu activated model

# Adapted stright from the the docs:
# https://lightning.ai/docs/pytorch/stable/notebooks/lightning_examples/mnist-hello-world.html#A-more-complete-MNIST-Lightning-Module-Example

class ReluLitMNISTWithAcc(LightningModule):
    def __init__(self, data_dir='.', hidden_size=64, learning_rate=2e-4):

        super().__init__()
        torch.manual_seed(42)

        # Set our init args as class attributes
        self.data_dir = data_dir
        self.hidden_size = hidden_size
        self.learning_rate = learning_rate

        # Hardcode some dataset specific attributes
        self.num_classes = 10
        self.dims = (1, 28, 28)
        channels, width, height = self.dims
        self.transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,)),
            ]
        )

        # Define PyTorch model
        self.model = nn.Sequential(
            nn.Flatten(),
            nn.Linear(channels * width * height, hidden_size),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, hidden_size),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, self.num_classes),
        )

        self.accuracy = Accuracy(task="multiclass", num_classes=10)

    def forward(self, x):
        x = self.model(x)
        return F.log_softmax(x, dim=1)

    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        preds = torch.argmax(logits, dim=1)
        self.accuracy(preds, y)

        # Calling self.log will surface up scalars for you in TensorBoard
        self.log("val_loss", loss, prog_bar=False)
        self.log("val_acc", self.accuracy, prog_bar=False)
        return loss

    def test_step(self, batch, batch_idx):
        # Here we just reuse the validation_step for testing
        return self.validation_step(batch, batch_idx)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



# Make a sigmoid activated model

# Adapted stright from the the docs:
# https://lightning.ai/docs/pytorch/stable/notebooks/lightning_examples/mnist-hello-world.html#A-more-complete-MNIST-Lightning-Module-Example

class SigmoidLitMNISTWithAcc(LightningModule):
    def __init__(self, data_dir='.', hidden_size=64, learning_rate=2e-4):

        super().__init__()
        torch.manual_seed(42)

        # Set our init args as class attributes
        self.data_dir = data_dir
        self.hidden_size = hidden_size
        self.learning_rate = learning_rate

        # Hardcode some dataset specific attributes
        self.num_classes = 10
        self.dims = (1, 28, 28)
        channels, width, height = self.dims
        self.transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,)),
            ]
        )

        # Define PyTorch model
        self.model = nn.Sequential(
            nn.Flatten(),
            nn.Linear(channels * width * height, hidden_size),
            nn.Sigmoid(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, hidden_size),
            nn.Sigmoid(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, self.num_classes),
        )

        self.accuracy = Accuracy(task="multiclass", num_classes=10)

    def forward(self, x):
        x = self.model(x)
        return F.log_softmax(x, dim=1)

    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        preds = torch.argmax(logits, dim=1)
        self.accuracy(preds, y)

        # Calling self.log will surface up scalars for you in TensorBoard
        self.log("val_loss", loss, prog_bar=False)
        self.log("val_acc", self.accuracy, prog_bar=False)
        return loss

    def test_step(self, batch, batch_idx):
        # Here we just reuse the validation_step for testing
        return self.validation_step(batch, batch_idx)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer


# Time and begin the process of assessing all test/train/validate splits
acc_list = list()
time_start = time.time()
for outer_fold, (outer_train_idx, outer_test_idx) in enumerate(kf_outer.split(selected_data)):
    
    # Perform the model runs
    
    torch.manual_seed(42)
    mnist_model_relu = ReluLitMNISTWithAcc()
    mnist_model_sigmoid = SigmoidLitMNISTWithAcc()
    print('outer_fold',outer_fold)
    print('outer_train_idx',len(outer_train_idx))
    outer_test_subset = Subset(selected_data,outer_test_idx)
    outer_val_dataset = DataLoader(outer_test_subset,batch_size=batch_size)
    print('outer_val_dataset',len(outer_val_dataset.dataset),len(outer_test_idx))
    print('outer_test_idx',len(outer_test_idx))
    print('')
    for inner_fold, (inner_train_idx, inner_test_idx) in enumerate(kf_inner.split(outer_train_idx)):
        print('inner_fold',inner_fold)
        print('inner_train_idx',len(inner_train_idx))
        inner_train_subset = Subset(selected_data,inner_train_idx)
        inner_train_dataset = DataLoader(inner_train_subset,batch_size=batch_size)
        print('inner_test_idx',len(inner_test_idx))
        inner_test_subset = Subset(selected_data,inner_test_idx)
        inner_test_dataset = DataLoader(inner_test_subset,batch_size=batch_size)

        # Make trainers
        if number_of_gpus == 0:
            trainer_relu = Trainer(accelerator="cpu",devices=number_of_cores,max_epochs=3,enable_progress_bar=False,deterministic=True,default_root_dir=log_path)
            trainer_sigmoid = Trainer(accelerator="cpu",devices=number_of_cores,max_epochs=3,enable_progress_bar=False,deterministic=True,default_root_dir=log_path)
        else:
            trainer_relu = Trainer(accelerator="gpu",devices=number_of_gpus,max_epochs=3,enable_progress_bar=False,deterministic=True,default_root_dir=log_path)
            trainer_sigmoid = Trainer(accelerator="gpu",devices=number_of_gpus,max_epochs=3,enable_progress_bar=False,deterministic=True,default_root_dir=log_path)

        # Work the magic: train and valide the models
        trainer_relu.fit(mnist_model_relu, inner_train_dataset)
        test_acc_raw_relu = trainer_relu.validate(mnist_model_relu, inner_test_dataset)
        test_acc_dict_relu = test_acc_raw_relu[0]
        val_acc_raw_relu = trainer_relu.validate(mnist_model_relu, outer_val_dataset)
        val_acc_dict_relu = val_acc_raw_relu[0]        
        acc_list.append(['relu',test_acc_dict_relu['val_acc'],val_acc_dict_relu['val_acc']])

        trainer_sigmoid.fit(mnist_model_sigmoid, inner_train_dataset)
        test_acc_raw_sigmoid = trainer_sigmoid.validate(mnist_model_sigmoid, inner_test_dataset, verbose=False)
        test_acc_dict_sigmoid = test_acc_raw_sigmoid[0]
        val_acc_raw_sigmoid = trainer_sigmoid.validate(mnist_model_sigmoid, outer_val_dataset, verbose=False)
        val_acc_dict_sigmoid = val_acc_raw_sigmoid[0]   
        acc_list.append(['sigmoid',test_acc_dict_sigmoid['val_acc'],val_acc_dict_sigmoid['val_acc']])
        
    print('')
    print('')

time_end = time.time()
elapsed_time = time_end-time_start
print('Elapsed time:')
print(elapsed_time)


# Use the processor / CPU / GPU metadata to output workflow metadata
df = pd.DataFrame(columns=["NumCPUs","NumGPUs","GPUType","CPUType","CPUSpeed","k","NumImages","Elapsed_Time","Metadata"],
                  data=[[cpu_count,number_of_gpus,gpu_type,cpu_type,cpu_speed,k,n_images,elapsed_time,output_string]])
df.to_csv("Time_Results/"+file_name+".csv",index=False)
pd.DataFrame(acc_list).to_csv("Accuracy_Results/accuracy_results_"+file_name+".csv",index=False)


# Consider: Clean up
# import shutil
# shutil.rmtree("MNIST")
# shutil.rmtree("lightning_logs")
