# Scientific Workflows Course

##  File Descriptions
This repository contains code to run a fully nested cross validation workflow on the TensorFlow "fashion_mnist" dataset of apparel images for the Science IT Scientific Workflows course with the goals described [below](#Goals). Descriptions of the files are as follows:

- The "README.md" is a standard file that should be added (as a best practice) to all code repositories.

- The ".gitignore" file is another standard file that can/should be added (as a best practice) to many code repositories. It ensures that unwanted files that exist within a Git repository are not tracked; e.g., checkpoints for Jupyter notebooks.

- The "TensorFlow_Tutorial_Jupyter.ipynb" file shows the analysis code within a Jupyter notebook (allowing for assessment of printed data objects and a better understanding of the code).

- The "TensorFlow_Benchmark_Results.ipynb" file shows a simple analysis of the timings that were recorded within the code. Note: this notebook uses a separate R kernel environment, which uses only 2 packages (`tidyverse` and `ggplot2`).

- The "TensorFlow_Tutorial_CommandLine.py" file is a ready-to-use Python 3 script for calling the code directly from a command line. The command line arguments are ordered as follows: (1) the number of cores to use; (2) the _k_ in k-fold cross validation; (3) the number of images to include in the analysis (currently limited to values between the specified _k_ value and 50,000)

- The "environment.yml" file details the packages that are used in the Conda environment that underlies the principal analytical code in this repository. It is detailed in the Conda section below.

- The "singularity.def" file details the creation of a Singularity environment that can also be used to run the analytical code. It is detailed in the Singularity section below.

- The "submit_array_job.sh" is a sample Slurm submission script that shows how to use [an array](https://docs.s3it.uzh.ch/cluster/parallelisation/#job-arrays) to execute multiple jobs within a "pleasingly parallel" workflow. More details on parallelization can be found in the Benchmarking section below.

- The "submit_slurm_jobs.sh" script is a Bash script that can be executed in the ScienceCluster via the `bash submit_slurm_jobs.sh` command. It submits the jobs that are defined in the Slurm submission scripts located in the "SlurmScripts" directory. This is an example of how you can submit multiple jobs at one time on the ScienceCluster using a single command.

- The "Accuracy_Results" and "Time_Results" directories are where the analytical accuracy ouputs and the timing outputs are stored, respectively.

- The "Images" directory is where the images for this repository (i.e., this README) are stored.

- The "Resource_Monitoring_and_Benchmarking" directory is a collection of materials for monitoring and benchmarking code.

- The "container_info.md" file has in-depth directions and guidance on using Singularity.

## Goals
In practice, the main analytical code in this repository would be written to compute a rigorous and unbiased accuracy estimate for a full model selection/training procedure. For Science IT purposes, this repository (and the analytical code contained within it) intends to accomplish the following goals:
- create an example analytical script that can be slightly altered for benchmarking purposes; e.g., the following aspects of the script can be adjusted to compare the speeds using various parallelization strategies:
    - the number of classified images
    - the number of cores
    - the presence/absence of a GPU
    - the _k_ in k-fold cross validation (i.e., a higher _k_ value results in a greater number of trained/validated/tested models)
- create a body of code that can be applied using either Anaconda and Singularity as an environment manager / creation tool to demonstrate multiple ways to create portable software environments
- sharing SLURM scripts that run the analytical code on Science IT ScienceCluster infrastructure to demonstrate the fundamentals of parallelization and benchmarking

---

## Installation / Preparing the Environment

### Conda
For beginners, we recommend using Anaconda (i.e., Conda) for installing the packages necessary to run this code. Using Anaconda will allow a user to run this code on the ScienceCloud, the ScienceCluster, and on ScienceApps.

To use Conda, create the environment before running any of the example scripts:
```
conda env create -f environment.yml
```

#### How was the `environment.yml` file created?

First, the underlying Conda environment should be created using the standard Conda commands; i.e., `conda create ...`, `source activate ENVNAME`, `conda install ...`. An example of a basic Conda environment creation can be found [here](https://docs.s3it.uzh.ch/how-to_articles/how_to_use_conda/#create-your-environment).

Once a Conda environment has been successfully created and refined, with all of the necessary packages, you can use the `conda list` or the `conda env export --name ENVNAME` command to list all package dependencies in the environment along with their specific versions. You can then use these details to build the `environment.yml` file, creating a new environment via `conda env create --file environment.yml` as demonstrated above. For reference, the original environment was built by manually installing (i.e., using `conda install ...` on) the following packages:
```
python=3.8.
tensorflow-gpu=2.2.0
scikit-learn
pandas
joblib
jupyterlab
ipykernel
pynvml
py-cpuinfo
matplotlib
ndcctools
```

To add this environment as a [custom kernel in Jupyter](https://docs.s3it.uzh.ch/cluster/apps/user_guide/#custom-kernels-in-jupyter) if you intend on using it on the ScienceApps, run the following code on a ScienceCluster terminal after creating the environment:
```
module load anaconda3
source activate tfbenchmark
ipython kernel install --user --name tfbenchmark
```

## Singularity (i.e., a Docker-style workflow that operates on shared cluster systems)

### First, some background:

#### What is a container? 

#### How does a container differ from a virtual machine?

### or from a virtual environment (such as conda or mamba)?

Singularity is another way to build your environment, allowing even greater portability across systems at different universities or computing centers.

###### [A short introduction to container concepts from Pawsey Supercomputing Centre](https://pawseysc.github.io/singularity-containers/11-containers-intro/index.html)  

### When should you use a container? Some possibilities:
When you are using ScienceCluster, but you have an application to build/install with many or complicated system dependencies (ubuntu packages).<br>
When you have an application that you wish to build and run on multiple systems.<br>
If someone has distributed an application as a docker or singularity container image.<br>

A full tutorial on using Singularity on ScienceCluster can be found [**here**](https://docs.s3it.uzh.ch/cluster/singularity_tutorial/).

Guide to the singularity definition file, for building your own custom singularity container:
[definition file](https://docs.sylabs.io/guides/3.5/user-guide/definition_files.html)

### Bonus example using OpenMPI parallelism

Containers using (MPI)[https://docs.sylabs.io/guides/3.5/user-guide/mpi.html] parallelism (Message Passing Interface) can be more complicated, and there are multiple ways to configure singularity plus MPI.

Here is an example using the **bind** approach.  A definition file could look like below.  It assumes that you have already compiled and built the executable mpi_hello_world from mpi_hello_world.c, borrowed from (here)[https://github.com/mpitutorial/mpitutorial/tree/gh-pages/tutorials/mpi-hello-world]:
```
Bootstrap: docker
From: ubuntu:20.04

%files
    ./mpi_hello_world /opt/bin/mpi_hello_world


%environment
    export MPI_DIR="/apps/opt/spack/linux-ubuntu20.04-x86_64/gcc-9.3.0/openmpi-4.1.3-hxw6m2yyxelnba6ruvojic4oohxsbdl7"
       ## ScienceCluster openmpi bind paths
    export PATH="$MPI_DIR/bin:$PATH"
    export LD_LIBRARY_PATH="$MPI_DIR/lib:$LD_LIBRARY_PATH"
    export LIBRARY_PATH="$MPI_DIR/lib:$LIBRARY_PATH"

%post
    echo "Installing base packages..."
    apt-get update -y  && . /etc/environment
    apt-get install hwloc -y
    apt-get install libnl-3-200 -y
    mkdir -p /opt/bin
```

In this mode, the application in the container calls the MPI library on the host (ScienceCluster) for communication between tasks. A slurm script to run this container using 4 MPI tasks on a single node could look like this:
```
#!/bin/bash
#SBATCH --ntasks=4  ## number of MPI tasks
#SBATCH --cpus-per-task=1  ## number of cores per task
#SBATCH --time=00:01:00
#SBATCH --mem 3000M 
#SBATCH --nodes=1
#SBTACH --ntasks-per-node=4


module load openmpi
module load singularityce

### run the container in bind mode

echo 'start'

srun --exclusive --ntasks=4 --cpus-per-task=1 --nodes=1  --mem 3000M singularity exec --bind "$MPI_DIR" --bind /apps ./mpi_hello_world /opt/bin/mpi_hello_world

echo 'finished'
```
Here is the same example using the **hybrid** approach and the definition file, hellompi_hybrid.def:
```
Bootstrap: docker
From: ubuntu:20.04

%files
    mpi_hello_world.c /opt
    makefile /opt

%environment
    export OMPI_DIR=/opt/ompi
    export SINGULARITY_OMPI_DIR=$OMPI_DIR
    export SINGULARITYENV_APPEND_PATH=$OMPI_DIR/bin
    export SINGULAIRTYENV_APPEND_LD_LIBRARY_PATH=$OMPI_DIR/lib

%post
    echo "Installing base packages..."
    apt-get update -y  && . /etc/environment
    apt-get install wget gcc bash g++ make libsysfs2 libsysfs-dev libevent-dev libpmix-dev -y

    echo "Installing Open MPI"
    export OMPI_DIR=/opt/ompi
    export OMPI_VERSION=4.1.3
    export OMPI_URL="https://download.open-mpi.org/release/open-mpi/v4.1/openmpi-$OMPI_VERSION.tar.bz2"
    mkdir -p /tmp/ompi
    mkdir -p /opt
    ##  Download
    cd /tmp/ompi && wget -O openmpi-$OMPI_VERSION.tar.bz2 $OMPI_URL && tar -xjf openmpi-$OMPI_VERSION.tar.bz2
    # Compile and install
    cd /tmp/ompi/openmpi-$OMPI_VERSION && ./configure --enable-orterun-prefix-by-default --with-pmix=/usr/lib/x86_64-linux-gnu/pmix/ --prefix=$OMPI_DIR && make install
    # Set env variables so we can compile our application
    export PATH=$OMPI_DIR/bin:$PATH
    export LD_LIBRARY_PATH=$OMPI_DIR/lib:$LD_LIBRARY_PATH
    export MANPATH=$OMPI_DIR/share/man:$MANPATH

    echo "Compiling the MPI application..."
    cd /opt
    make
```
Notice that Open MPI is installed directly into the container. The files 'mpi_hello_world.c', 'makefile', and hellompi_hybrid.def would be in the current working directory.  Because of the sudo requirement, you would need to create the container on a ScienceCloud virtual machine (we have images with singularity pre-installed) or on your own machine. Then the command to build the container image is:
```
sudo singularity build hellompi_hybrid.sif hellompi_hybrid.def
```
You would then transfer the .sif image file to ScienceCluster.  Then, after expanding the image with:
```
singularity build --sandbox ./hellompi_hybrid ./hellompi_hybrid.sif
```
The executable within the image can be run with the following slurm script:
```
#!/bin/bash
#SBATCH --ntasks=4  ## number of MPI tasks
#SBATCH --cpus-per-task=1  ## number of cores per task
#SBATCH --time=00:01:00
#SBATCH --mem 3000M 
#SBATCH --nodes=1
#SBTACH --ntasks-per-node=4


module load openmpi/4.1.3
module load singularityce


echo 'start'

srun singularity exec ./hellompi_hybrid  /opt/mpi_hello_world

echo 'finished'
```


## Prepare the tensorflow benchmark using singularity

To use Singularity, create the Singularity Image file from the definition file using:
```
sudo singularity build tfbenchmark.sif singularity.def
```
Then transfer the `tfbenchmark.sif` file to your system of interest and sandbox it:
```
singularity build --sandbox tfbenchmark tfbenchmark.sif
```
After that, use either `singularity shell` or `singularity exec` to run the code (i.e., integrate these commands with a sample run, such as `python TensorFlow_Tutorial_CommandLine.py 3 5 500 tfb`)

There are ScienceCloud images that come equipped with Singularity pre-installed; i.e., the image titled "***Singularity 3.8 Ubuntu 20.04 (2021-07-06)" that's selectable from the "Source" menu of the ScienceCloud dashboard "Launch Instance" menu.

---
**NOTE**

It is also possible to create a Conda environment within a Singularity container. This tutorial demonstrates, however, how to make a slightly different environment—with the same packages required to run the production code, just different versions—in Singularity. (Only the `ndcctools` package was omitted so as to keep the tutorial as simple as possible.)

---

For reference: further installation instructions for TensorFlow, including platform-specific and GPU instructions, can be found [here](https://www.tensorflow.org/install/).


## Running the code from the command line

Once the software environment has been created, it will be possible to run the analysis code directly from the command line. To do so, activate your chosen environment (e.g., `source activate tfbenchmark` on the ScienceCluster) then run:

```
python TensorFlow_Tutorial_CommandLine.py 3 5 500 tfb
```

As currently implemented, the command line code requires 4 arguments; in order, they are:

1. the number of cores to use
2. `k` for fully nested K-Fold CV
3. the number of images to keep from the full dataset
4. a metadata string (usually a single word) to organize output files

Suggested values for arguments 1-3 when running on a personal computer are 3, 5 and 500, which takes approximately 15 seconds (on a 2018 MacbookPro).

The current maximum accepted value for the third argument is `50000` (as this is the number of images available in the current dataset for classification).

The code will produce a time file and an accuracy file with all inputted arguments included in its file name in addtion to the date/time the file was produced. These files will be delivered to their respective directories.

---
**NOTE**

The first time the code runs, it needs an internet connection to download training/testing datasets.

---

# Advanced Training Workflows Course Notes

## Code Description

As noted above, this code demonstrates a nested K-Fold cross validation example that assesses the generalized accuracy of 2 TensorFlow based models that operate on the ["fashion_mnist" dataset](https://www.tensorflow.org/datasets/catalog/fashion_mnist).

### What is k-fold cross validation?

![K-fold CV](Images/K-fold_cross_validation_EN.svg)

[K-Fold Cross validation](https://en.wikipedia.org/wiki/Cross-validation_(statistics)) is a method that splits a dataset into k-pieces that are then rotated between test sets and training sets. The idea is to minimize potential bias in any single test dataset so as to best generalize how a model or statistical analysis will generalize to an independent dataset.

While K-fold CV can be used as a method to generalize the results of a statistical model on an independent dataset, it is also commonly used as a method to select the "best model" from a selection of potential models (either choosing the "best" hyperparameter set for a model and/or comparing completely different model types). However, when K-Fold CV is used both for accuracy assessment as well as model selection, bias/"data leakage" can occur. A useful starting point for reading more about these discussions is [this paper](https://www.sciencedirect.com/science/article/pii/S0957417421006540).

To reduce the effects of such bias, nested K-Fold cross validation can be used instead. In this method, a dataset is partitioned into k-folds that are then each used as train, validate, and test sets. This results in many more model iterations than with a "flat" CV procedure.

A nested K-Fold CV procedure was selected as an example because it allows us to experiment with various types of parallelization, across multiple CPUs and across multiple GPUs, as well as with separate tasks and within the code itself.

## Monitoring and Benchmarking

The process of benchmarking can be [defined](https://en.wikipedia.org/wiki/Benchmark_(computing)) as "the act of running a computer program, a set of programs, or other operations, in order to assess the relative performance of an object, normally by running a number of standard tests and trials against it." Refer to the ["Resource_Monitoring_and_Benchmarking" directory](Resource_Monitoring_and_Benchmarking/README.md) for a full suite of tools to reference when monitoring and benchmarking code.

## Parellelization: What is it exactly and what options do I have?

Commonly, you can consider 2 basic strategies for parallelizing your work:

1. parallelization within the analytical code itself (either across multiple CPUs or GPUs); this is sometimes called ["multithreading"](https://docs.s3it.uzh.ch/cluster/parallelisation/#multithreaded-application)
2. parallelization at the task level

Depending on the details of your own analytical code, one or both types of parallelism may be possible. Only *you*, the domain expert and the author of your own code, will be able to make the most informed decision about parallelism in your workflow.

#### Multithread / Multi-GPU parallelization

The first sort of parallelization is specific to the analytical code. In this repository, such parallelization is illustrated in the "TensorFlow_Tutorial_Jupyter.ipynb". Within this notebook, you can see that the `Parallel` function from the `joblib` library is used to spread the multiple test/validate/test permutations across multiple CPUs.

Moreover, because this code uses Tensorflow, the `strategy = tf.distribute.MirroredStrategy()` option can be used to parallelize the data across multiple GPUs.

Notice that these 2 methods of parallelization are specific to the software language (python), the tools being used (Tensorflow models), and the analysis being run (multiple iterations of models with different train/validate/test sets). In other words, implementing this sort of parallelism requires a knowledge of the code itself and may not be possible depending on the analysis and/or the tools being used.

##### Monitoring GPU Usage

If you would like to monitor GPU usage, you can similarly execute the analytical code in a background terminal/process then run the following command in another terminal that's part of the same session (i.e., located on the same computer or node):
```
nvidia-smi -i $CUDA_VISIBLE_DEVICES -l 2 --query-gpu=gpu_name,memory.used,memory.free,timestamp --format=csv -f nvidia-smi.log
```

In this case, the `nvidia-smi` tool will record all details on the devices that are assigned via `$CUDA_VISIBLE_DEVICES`, sampling data every 2 seconds (via the `-l 2` flag), returning the `gpu_name` and memory details, then writing all details in a `csv` style format to a file titled `nvidia-smi.log`.



#### Task/Job Parallelization

The second sort of parallelization occurs at the "task" or "job" level. Because this code can be executed with multiple parameters (e.g., for `k` in k-fold cross validation), multiple executions of the code can be run completely independently of one another. Parameter sweeps are commonly able to take advantage of this type of parallelization.

This type of circumstance has been termed ["embarrasingly parallel"](https://en.wikipedia.org/wiki/Embarrassingly_parallel). It is "embarrassingly", "delightfully", or "pleasingly" parallel because none of the executions of the code require any interaction with the other executions of the code. This means that in a large enough computational environment *all* iterations of the code can be run potentially simultaneously. Such a workflow can be accomplished on cluster type systems using the array framework demonstrated in the "submit_array_job.sh" script template.

Furthermore, you can parallelize runs with multiple array jobs on different partitions of a large cluster type environment (i.e., the ScienceCluster) by queuing multiple jobs across multiple partitions simultaneously, as illustrated with the "submit_slurm_jobs.sh" script. In this context, all independent array jobs will run as soon as resources are available for them.

### How do I determine the optimal number of cores / GPUs to use?

If you can indeed perform parallelization at the code level (i.e., a multithreaded application or multi-GPU application), the next question then becomes, "what level of parallelization will be most optimal?"

Common sense might dictate that the greater number of cores or GPUs that you use, the faster your workflow will become. However, per Amdahl's law, this is not the case.

![Amdahl's Law](Images/1920px-AmdahlsLaw.svg.png)

In technical terms, [Amdahl's law](https://en.wikipedia.org/wiki/Amdahl%27s_law) states that "the overall performance improvement gained by optimizing a single part of a system is limited by the fraction of time that the improved part is actually used". Otherwise stated, Amdahl's law means that you can only parallelize your workflow up to a certain maximal point of efficiency (unique to the application / analysis code itself), after which increased resources provided for parallelization (i.e., adding more CPUs or GPUs) will not result in any greater efficiency. In reality, after reaching the "optimal threshold", greater increases in provided resources may actually **signficantly decrease** the efficiency of the code's execution. This dropoff in efficiency can be seen in the SpeedUp charts within the "TensorFlow_Benchmark_Results.ipynb" notebook.

In order to find the optimal level of hardware for efficiency gains, you will simply need to test multiple levels of provided hardware and chart/log the efficiency as resources increase. For this context, efficiency is measured as "speedup", which is the non-parallelized implementation time divided by the parallelized implementation time (see the use of the `mutate` function in cell 5 of the "TensorFlow_Benchmark_Results.ipynb" notebook). After the "sweet spot" of parallelization is reached, no greater efficiency can be accomplished (but efficiency can greatly decrease).

**TL;DR**: a greater number of cores or GPUs **does not** mean an automatic linear increase in efficiency. You will need to perform a benchmarking process to determine where the "sweet spot" for parallelization exists.
