# Intro to Monitoring, Benchmarking, and Profiling

[[_TOC_]]

## Why?

A common misconception is that requesting more hardware will automatically speed up our code. But slow code is often related to the code itself, so what about optimizing our code? But as you might know

> premature optimization is the root of all evil

This is the famous quote by Donald Knuth:

> Programmers waste enormous amounts of time thinking about, or worrying about, the speed of noncritical parts of their programs, and these attempts at efficiency actually have a strong negative impact when debugging and maintenance are considered. We should forget about small efficiencies, say about 97% of the time: premature optimization is the root of all evil.

> Yet we should not pass up our opportunities in that critical 3%.

Source: [Structured Programming with _go to_ Statements](https://doi.org/10.1145/356635.356640)

So we should try to identify the critical parts, called also **bottlenecks**. Improving the runtime performance is an iterative process:

1. Find the bottlenecks
2. Try to eliminate them (optimize, parallelize)
3. Repeat until you are happy

Moreover, with Science IT computing infrastructure, the cost contribution is based on a "pay-per-use" (known also as "pay-as-you-go") model, where by "use" means reserved or allocated resources, and not what is really in usage during the computational work. So once we are happy with the code performance, we should try to optimize the cost by using a sensible amount of resources.

The aim of this guide is to introduce techniques to that will help to:

- measure the total execution time of your application
- understand your code performance by examining the system utilization (CPU and Memory)
- find the bottlenecks by measure the runtime and memory usage of an isolated piece of code (Python/R)
- avoid common pitfalls in assessing the performance on various systems
- use Slurm to understand the resource usage of your jobs (running or completed)
- optimize the cost

## Benchmarking

Benchmarking the code means evaluating the execution time.

The first step is to measure the execution time of the whole application.

### Bash `time`

Info: There is another `/usr/bin/time`

```bash
$ time python example_1.py

real	0m2.096s
user	0m0.012s
sys 	0m0.010s
```

- **real**: the elapsed time (wall clock) and is the time used by the process including also input/output (IO) operations .
- **user**: CPU time spent in user-mode code (outside the kernel) within the process (it includes all cores)
- **sys**: CPU time spent in the kernel with the process

```bash
$ time python example_2.py

real    0m13.369s
user    1m30.687s
sys     0m3.319s
```

When benchmarking:

- make sure the system is not used for other tasks (we can check this by monitoring the resources)
- it is hard to compare different infrastructures (we should avoid doing it)
- it is hard to compared when using different inputs (we should avoid doing it)
- IO operations make the benchmarking less predictable (more later)

## Resource Monitoring (CPU and Memory)

First we want to understand the utilization of the system and what processes are active.
Each operating system has established tools for this:

- [Resource Monitor](https://en.wikipedia.org/wiki/Resource_Monitor) on Windows
- [Activity Monitor](https://en.wikipedia.org/wiki/List_of_built-in_macOS_apps#Activity_Monitor) on macOS
- [htop](<https://en.wikipedia.org/wiki/Top_(software)>) on UNIX-like systems (including ScienceCloud and ScienceCluster)

### `htop`

[`htop`](https://manpages.ubuntu.com/manpages/latest/en/man1/htop.1.html) is a real-time interactive process viewer.

(if not available please install it, e.g., `$ brew install htop` on macOS)

```bash
# for all processes
$ htop

# to exit press F10 or use the cursor

# for my processes
$ htop -u $USER
```

htop is interactive, so we can use the cursor for various tasks, e.g., sort the processes by CPU%.

What about the unstable internet connection?

### `screen`

It can be used to resume a session.

You can create sessions, detach, reattach, and terminate them.

```bash
$ screen -S <name> # create a named session and attach to it

Ctrl+a d # detach from the session

$ screen -r # reattach to the session (if you have multiple sessions you need to give the pid)
$ screen -ls #list all sessions
$ screen -r <pid> # reattach to the session with given pid

$ exit # the session terminates when you exit (run it in the session)
```

More info at https://linuxize.com/post/how-to-use-linux-screen/ .

## Finding the Bottlenecks

Adding timestamps in our code to identify the bottleneck can be tedious and time consuming. We should should rather use dedicated tools able to measure the instructions or memory usage at the function or code line level. This activity is called **profiling** and the tools are know as **profilers**.

Profiling is introducing a significant run time overhead, so do not benchmark your code when profiling. Once the bottleneck is identified, we should try to isolate the corresponding code section and perform more benchmarking (or micro-benchmarking).

Profiling parallel code is hard.

### Python

- logging: [`logging`](https://docs.python.org/3/howto/logging.html) module (see the `DEBUG` level)
- micro-benchmarking: [`timeit`](https://docs.python.org/3/library/timeit.html) allows you to time small pieces of code.
- profiling [`cProfile`](https://docs.python.org/3/library/profile.html) might be intimidating. For more examples, check out this [blog post](https://realpython.com/python-profiling/).
- memory profiling: [`memray`](https://github.com/bloomberg/memray)

### R

- If you are using **R**, check out Hadley Wickham's [post about optimizing code](http://adv-r.had.co.nz/Profiling.html).

If you are using **RStudio**, check out the [`profvis`](https://rstudio.github.io/profvis/) package.

```R
library(profvis)

profvis({
  data(diamonds, package = "ggplot2")

  plot(price ~ carat, data = diamonds)
  m <- lm(price ~ carat, data = diamonds)
  abline(m, col = "red")
})
```

### JupyterLab

Basic resource monitoring and Python profiling can be done directly in JupyterLab. The main packages that you need are available in [environment.yml](environment.yml).

Useful info also at [Profiling and Timing Code by Jake VanderPlas (2016)](https://jakevdp.github.io/PythonDataScienceHandbook/01.07-timing-and-profiling.html). Be aware that the[`memory_profiler`](https://github.com/pythonprofilers/memory_profiler) is no longer actively maintained.

Now we will use directly [Python.ipynb](Python.ipynb).

**What is faster:**

```{python}
import numpy as np

n = 10
sum(range(n))
np.arange(n).sum()
```

Allocating memory makes the code slow. Why?

## Computer Architecture

We have to start with the computer architecture (see below basic representation that might not be 100% accurate).

```mermaid
flowchart LR
    CPU <--> IO

subgraph CPU

    Core_1(Core 1)
    Core_n(Core n)
    Core_1 --> Level_2_Cache(Level 2 Cache)
    Core_n --> Level_2_Cache
    Level_2_Cache --> Level_3_Cache(Level 3 Cache)
    Level_3_Cache --> Main_Memory("Main Memory (RAM)")
end

subgraph Core_1
    direction LR
    Registers_1(Registers)
    Arithemetic_Unit_1(Arithmetic Unit)
    Control_Logic_1(Control Logic)
    Level_1_Cache_1(Level 1 Cache)
end

subgraph Core_n
    direction LR
    Registers_n(Registers)
    Arithemetic_Unit_n(Arithmetic Unit)
    Control_Logic_n(Control Logic)
    Level_1_Cache_n(Level 1 Cache)
end

subgraph IO
    direction LR
    Storage("Storage(e.g., SSD, Disk)")
    Network
    ...

end
```

Latency numbers (2020 values):

```
1 CPU Cycles (~3 GHz) ...................... 0.3 ns
L1 cache reference ........................... 1 ns
L2 cache reference ........................... 4 ns
Main memory reference ...................... 100 ns
Send 1KB over 10 Gb/s network ............ 1_000 ns  =   1 µs
SSD random read ......................... 16_000 ns  =  16 µs
Read 1 MB sequentially from memory ....... 1_000 ns  =   1 µs
Round trip within same datacenter ...... 500_000 ns  = 500 µs
Read 1 MB sequentially from SSD ......... 19_000 ns  =  19 µs
Disk seek ............................ 2_000_000 ns  =   2 ms
Read 1 MB sequentially from Disk ....... 474_000 ns  = 474 µs
Send packet CA->Netherlands->CA .... 150_000_000 ns  = 150 ms
```

Credit:

- https://people.eecs.berkeley.edu/~rcs/research/interactive_latency.html
- By Jeff Dean: http://research.google.com/people/jeff/
- Originally by Peter Norvig: http://norvig.com/21-days.html#answers

The high level languages (e.g., Python, R) are not designed to control the cache usage.

In practice we can reduce the execution time by paying attention memory usage and IO operations. We should copy the data close to the compute infrastructure (or even in memory).

## Job monitoring with Slurm

The job scheduler Slurm can report statistics for jobs that can help you understand what resources to request. Here is a short guide to the steps for resource monitoring.

_note_: Slurm makes measurements every 60 seconds, which may result in instantaneous spikes not being measured.

_hint_: [Resource monitor](https://cctools.readthedocs.io/en/latest/resource_monitor/) from CCTools is very useful as it can track most subprocesses and has a relatively small overhead. It also tracks much more than Slurm does.

### Estimating your hardware requirements

It works well for embarrassingly parallel problems.

**Sample!** Take a representative small sample from a batch, whether it's data, file or parameters. If the batch is heterogeneous, take a few representative samples. If possible, run a pipeline or workflow start-to-finish with that sample. Run the same pipeline with more resources and ask yourself, "does it speed up?" Extrapolate the time needed for the entire data or parameter set from this small sample. Doing this will also help you find bugs or errors in the code much faster than when starting with the entire data or parameter set.

### Running jobs

For currently running jobs, you can use [`squeue`](https://slurm.schedmd.com/squeue.html), and specifically for your own jobs: `squeue -u $USER` or `squeue --me`. The output of this command shows the job ID in the first column (for running jobs only).

With the job ID for your currently running job, you can run [`sstat`](https://slurm.schedmd.com/sstat.html).

```bash
$ sstat --all --format JobID,NTasks,MaxRSS,MinCPU,AveCPU -j <jobid>
```

_hint_: add [`watch`](https://manpages.ubuntu.com/manpages/latest/en/man1/watch.1.html) before `sstat` for automatic refreshing

While a job runs you can also use `scontrol` to check what resources Slurm has allocated along with other important details of your job.

```bash
$ scontrol show job <jobid>
```

### Completed jobs

To find job IDs for jobs that have stopped running, use [`sacct`](https://slurm.schedmd.com/sacct.html), and in this example, my user's jobs from the last 30 days:

```bash
$ sacct -u $USER -S `date --date="-30 days" +%Y-%m-%d` -X -o jobid,jobname,start,end,elapsed,state
```

From the first column, select a job ID for a job. A job that ran for longer may be easier to understand, so if you have one, choose a job with a longer elapsed time.

```bash
$ sacct --format=jobid,ReqMem,MaxRSS,MaxVMSize,Elapsed,AllocCPU,SystemCPU,UserCPU -j <jobid>
```

The value for `CPUTime` is `NCPUS * Elapsed`

_hint_: `sacct -e` shows the available output columns.
