# Basics of Parallelization / GPU & Multi-GPU Usage

## Table of contents

1. [ What is parallelization? ](#what)
2. [ Why use parallel computing? ](#why)
3. [Hardware for parallel computing](#hardware)
4. [Understand the Problem and the Program](#program)
    1. [Check resources](#resources)
    2. [Monitoring GPU Usage](#monitoring)
    3. [Task/Job Parallelization](#jobarray)
    4. [How do I determine the optimal number of cores / GPUs to use?](#amdahl)


<a name="what"></a>

## What is parallelization?

> By parallelization it is meant performing parallel computing, i.e., the **simultaneous** use of multiple compute resources to solve a computational problem.
> 
> * A problem is broken into discrete parts that can be solved concurrently
> * Each part is further broken down to a series of instructions
> * Instructions (tasks) from each part execute simultaneously on different processors
> * An overall control/coordination mechanism is employed
>
> The compute resources are typically:
> 
> - A single computer with multiple processors/cores or accelerators like GPUs
> - An arbitrary number of such computers connected by a network

We will see an example of single-node parallelization, across *multiple CPU cores or multiple GPUs*, a typical case where communication between tasks is needed: the data or task information has then to be gathered and at the same time summarized. 

There are special cases of parallelization that require little to no need for coordination between the tasks, known as *embarrassingly parallel*. Examples are problems solved by simultaneously executing similar, but independent tasks many times (dozens to millions of times), e.g. identical piece of code to be run on different datasets or different set of parameters. 


<a name="why"></a>

## Why use parallel computing?

> Compared to serial computing, parallel computing is much better suited for modeling, simulating and understanding complex, real world phenomena.
>
>  Main reasons for using parallel computing:
>
> - we want to solve the same problem faster
> - we need to solve a larger problem, which doesn't fit anymore on the computer we are using
> - we have a very large number of such problems that we need to solve, which cannot fit into one computer

Each reason has some challenges that we need to address.

Depending on the details of your own analytical code, parallelism may be possible. Only *you*, the domain expert and the author of your own code, will be able to make the most informed decision about parallelism in your workflow.


<a name="hardware"></a>

## Hardware for parallel computing

>  CPU and GPU difference at hardware level:
> 
> |              | CPU                                                                               | GPU                                                        |
> |:------------:|:---------------------------------------------------------------------------------:|:----------------------------------------------------------:|
> | **Function** | Generalized component that deals with the main processing functions of a computer | Specialized component that is great for parallel computing |
> | **Emphasis** | Low latency                                                                       | High throughput                                            |
> | **Use case** | General purpose computing                                                         | Repetitive and high parallel computing tasks               |
> 
> 

 

A CPU can never be fully replaced by a GPU: *GPU complements CPU architecture*.

Be aware that transferring data to and from the GPU can be a bottleneck, and that for simple/small models it may take more time to send the code to the GPUs and back than running the code itself. The use of GPUs is not efficient in those cases!

When multi nodes are involved in the computation, data has to travel from one node to another node through the network. Nodes having *Infiniband* connections (low latency) are therfore ideal for running multi-node jobs.
The use of **single-node** or **multi-node parallelism** depends on if the program might benefit from shared memory or the distributed one. It mostly depends on IOs and network, i.e. if a program does a lot of IOs on a disk we may want to write from multiple nodes rather than all the processes from the same node on the same disk. 

More information about the ScienceCluster hardware can be found under [Cluster Resources](https://docs.s3it.uzh.ch/cluster/resources/#resources) in the Science IT docs.

<a name="program"></a>

## Understand the Problem and the Program

> Before spending time in an attempt to develop a parallel solution for a problem determine whether or not the problem is one that can actually be parallelized.
> 
> Then, in practice we need a programming model, which is a library or setup that allow to write the program so that is starts to work in parallel.
> 
>|                               | `Lightning_Tutorial_Jupyter.ipynb`                                                                                                                                                                                              |
> |-----------------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
> | **The problem**               | Nested K-Fold cross validation example: a dataset is partitioned into k-folds that are then each used as train, validate, and test sets. This results in multiple iterations of models with different train/validate/test sets. |
> | **Tool**                      | PyTorch Lightning                                                                                                                                                                                                               |
> | **Method of parallelization** | the `Trainer` function from the `pytorch_lightning` library                                                                                                                                                                     |
>


The parallelization is in general not *automagic*, it can be if the library or the program you are using has been designed for it.

Within the `Lightning_Tutorial_Jupyter.ipynb` notebook, the `Trainer` function from the `pytorch_lightning` library is used to spread the multiple train/validate/test permutations across multiple CPUs.
Moreover, the Trainer will run on all available GPUs by default (`accelerator="auto"`) .

```bash
   # run on one GPU
   trainer = Trainer(accelerator="gpu", devices=1)
   # run on multiple GPUs
   trainer = Trainer(accelerator="gpu", devices=8)
```

Note that this method of parallelization is specific to the tools being used (PyTorch Lightning), and the analysis being run (multiple iterations of models with different train/validate/test sets). In other words, implementing this sort of parallelism requires a knowledge of the code itself and may not be possible depending on the analysis and/or the tools being used.

<a name="resources"></a>

### Check resources

For a program to start working in parallel on multiple CPU cores or GPUs, the required resources need to be available.

In SLURM submission scripts:

|                        | Multi-CPU cores           | Multi-GPUs                     |
|------------------------|:-------------------------:|:------------------------------:|
| **In this repository** | `submit_multicpus_job.sh` | `submit_multigpus_job.sh`      |
| **Resources**          | N CPUs on the same node   | N GPUs                         |
| **SLURM option**       | `--cpus-per-task=N`       | `--gpus=N`  or `--gpus=type:N` |

Submit the submission scripts:

```bash
sbatch submit_multicpus_job.sh
module load v100
sbatch submit_multigpus_job.sh
```

See S3IT docs for [available GPUs resources] (https://docs.s3it.uzh.ch/cluster/resources/#gpus ) and the corresponding modules to load.

Information about allocated resources for the job can be obtained with the `scontrol` or the `sacct` command:

```bash
scontrol show <jobid>
sacct --format Jobid,ReqMem,MaxRSS,AllocCPUS,CPUTime,TotalCPU,"AllocTRES%-42"
```

where `CPUTime` is the time allocated to the job (`Elapsed*AllocCPUs`) and `TotalCPUs is the actual CPU time consumed by the job.

Refer to the ["Resource\_Monitoring\_and\_Benchmarking" directory](/devin.routh/scientific-workflows/-/blob/master/Resource_Monitoring_and_Benchmarking/README.md) for a full suite of tools to reference when monitoring and benchmarking code.

<a name="monitoring"></a>

### Monitoring GPU Usage

> The command `nvidia-smi` monitors and manages NVIDIA GPUs. It returns information about:
> 
>- all available GPUs
>- processes using GPUs
>
> Monitoring approaches:
> 
>  - schedule an [interactive session](https://docs.s3it.uzh.ch/how-to_articles/how_to_run_an_interactive_session/) with `srun`
>  - launch screen on the compute node or start the program in the background and monitor in the foreground

When submitting a program in batch mode, monitoring possibilities are somewhat limited, i.e commands like `top` and `ps` would show processes that run on the login node rather than on the node where the program is running. For monitoring purposes using [interactive session](https://docs.s3it.uzh.ch/how-to_articles/how_to_run_an_interactive_session/) is needed.

After connecting from a terminal to the ScienceCluster, work through the following steps:

```bash
    # load the gpu module

    module load v100

    # request an interactive session, which allows the package installer to see the GPU hardware

    srun --pty -n 1 -c 3 --time=01:00:00 --gpus=v100:2 --mem=8G bash -l

    # (optional) confirm the gpu is available. The output should show basic information about at least
    # one GPU.

    nvidia-smi

    # activate the virtual environment

    source activate tfbenchmark
```

#### Monitoring using screen

 When [`screen`](https://www.gnu.org/software/screen/manual/screen.html#Invoking-Screen) is called, it creates a single window with a shell in it (or the specified command) and you can use the program as you normally would:

```bash
# create a named session and attach to it
screen -S <name> 

# In the screen session
# create new (full-screen) windows with a shell and switch to that window
Ctrl+a c
# switch the input focus to the next region (windows)
Ctrl+a Tab
```

```bash
# on the first screen windows run
python Lightning_Tutorial_Jupyter.py 3 5 100000 tfb
# on the second windows run nvidia-smi
nvidia-smi -l 2
```


Then, at any time, you can:
```bash
# detatch from the session
Crtl+a c
# list all sessions
screen -ls
# delete a screen session
screen -S <pid> -X quit
```

Refer to the ["Resource\_Monitoring\_and\_Benchmarking" directory](/devin.routh/scientific-workflows/-/blob/master/Resource_Monitoring_and_Benchmarking/README.md) for more information about using `screen` for resource monitoring.

#### Monitoring by starting the program in background


You can similarly execute the analytical code in a background process then run the `nvidia-smi` command in the foreground.

```bash
python Lightning_Tutorial_Jupyter.py 3 5 100000 tfb &
nvidia-smi -i $CUDA_VISIBLE_DEVICES -l 2 --query-gpu=gpu_name,memory.used,memory.free,timestamp --format=csv -f nvidia-smi.log
```

In this case, the `nvidia-smi` tool will record all details on the devices that are assigned via `$CUDA_VISIBLE_DEVICES`, sampling data every 2 seconds (via the `-l 2` flag), returning the `gpu_name` and memory details, then writing all details in a `csv` style format to a file titled `nvidia-smi.log`.

<a name="jobarrays"></a>

### Task/Job Parallelization


> The second sort of parallelization occurs at the "task" or "job" level:
> 
> - the `Lightning_Tutorial_Jupyter.py` can be executed with multiple parameters (e.g., for `k` in k-fold cross validation)
> - multiple executions of the code can be run completely independently of one another. 

Parameter sweeps are commonly able to take advantage of this type of parallelization.
This type of circumstance has been termed ["embarrassingly parallel"](https://en.wikipedia.org/wiki/Embarrassingly_parallel). It is "embarrassingly", "delightfully", or "pleasingly" parallel because none of the executions of the code require any interaction with the other executions of the code. This means that in a large enough computational environment *all* iterations of the code can be run potentially simultaneously. 

In this repository, such parallelization is illustrated in the "submit\_array\_job.sh" script template.
 
Furthermore, you can parallelize runs with multiple array jobs on different partitions of a large cluster type environment (i.e., the ScienceCluster) by queuing multiple jobs across multiple partitions simultaneously, as illustrated with the "submit\_slurm\_jobs.sh" script. In this context, all independent array jobs will run as soon as resources are available for them.

#### Slurm job arrays

Such a workflow can be accomplished on cluster type systems using [job arrays in Slurm](https://docs.s3it.uzh.ch/cluster/parallelisation/#job-arrays). Job arrays offers a mechanism for submitting and managing collections of similar jobs quickly and easily: each job is submitted separately (*task-level parallelism* ) but the array of jobs can be managed as a whole. 

 When using the  `--array=1-N` Slurm option, N CPUs are requested to launch N completely independent jobs. Slurm will create N jobs, all identical except for the environmental variable `$SLURM_ARRAY_TASK_ID`: all jobs have the same jobID with an underscore

Note that, the `$SLURM_ARRAY_TASK_ID` is an integer. Use bash array when you have not integer parameters:

```
#SBATCH --array=0-2

FILES=(data*csv)

./myprog ${FILES[$SLURM_ARRAY_TASK_ID]}

or

PARAMS=($(seq 0 0.1 1))
./myprog ${PARAMS[$SLURM_ARRAY_TASK_ID]}
```

More information about [job arrays](https://docs.s3it.uzh.ch/cluster/parallelisation/#job-arrays) can be found in the Science IT docs.

 <a name="amdahl"></a>
 
### How do I determine the optimal number of cores / GPUs to use?

> If you can indeed perform parallelization at the code level (i.e., a multi-CPU cores application or multi-GPU application), the next question then becomes, "what level of parallelization will be most optimal?"
> 
> Common sense might dictate that the greater number of cores or GPUs that you use, the faster your workflow will become. However, per Amdahl's law, this is not the case.
> 
> ![Amdahl's Law](Images/1920px-AmdahlsLaw.svg.png)
> 
> 
> -  **speedup** (measure of the performance improvement) is the ratio between the time of a serial execution by by the time taken by the same operations executed in parallel
> - **efficiency** is the ratio between the speedup and the number of processors used


In technical terms, [Amdahl's law](https://en.wikipedia.org/wiki/Amdahl%27s_law) states that "the overall performance improvement gained by optimizing a single part of a system is limited by the fraction of time that the improved part is actually used". Otherwise stated, Amdahl's law means that you can only parallelize your workflow up to a certain maximal point of efficiency (unique to the application / analysis code itself), after which increased resources provided for parallelization (i.e., adding more CPUs or GPUs) will not result in any greater efficiency. In reality, after reaching the "optimal threshold", greater increases in provided resources may actually **significantly decrease** the efficiency of the code's execution. This dropoff in efficiency can be seen in the Speedup charts within the `TensorFlow\_Benchmark\_Results.ipynb` notebook.

In order to find the optimal level of hardware for efficiency gains, you will simply need to test multiple levels of provided hardware and chart/log the efficiency as resources increase. For this context, efficiency is measured as "speedup", which is the non-parallelized implementation time divided by the parallelized implementation time (see the use of the `mutate` function in cell 5 of the "TensorFlow\_Benchmark\_Results.ipynb" notebook). After the "sweet spot" of parallelization is reached, no greater efficiency can be accomplished (but efficiency can greatly decrease).

Reasons can be multiple but often is that:

- not all part of the code can be parallelized
- parallel tasks require time known as **parallel overhead**
- load imbalance between the computing units that execute tasks and process data.


**TL;DR**: a greater number of cores or GPUs **does not** mean an automatic linear increase in efficiency. You will need to perform a benchmarking process to determine where the "sweet spot" for parallelization exists.

